'use strict';

Object.defineProperty(exports, "__esModule", {
  value: true
});

var _slicedToArray = function () { function sliceIterator(arr, i) { var _arr = []; var _n = true; var _d = false; var _e = undefined; try { for (var _i = arr[Symbol.iterator](), _s; !(_n = (_s = _i.next()).done); _n = true) { _arr.push(_s.value); if (i && _arr.length === i) break; } } catch (err) { _d = true; _e = err; } finally { try { if (!_n && _i["return"]) _i["return"](); } finally { if (_d) throw _e; } } return _arr; } return function (arr, i) { if (Array.isArray(arr)) { return arr; } else if (Symbol.iterator in Object(arr)) { return sliceIterator(arr, i); } else { throw new TypeError("Invalid attempt to destructure non-iterable instance"); } }; }();

var _events = require('events');

var _errors = require('./errors');

var _indices = require('./indices');

var _attributes = require('./attributes');

var _edges = require('./iteration/edges');

var _neighbors = require('./iteration/neighbors');

var _serialization = require('./serialization');

var _utils = require('./utils');

function _toConsumableArray(arr) { if (Array.isArray(arr)) { for (var i = 0, arr2 = Array(arr.length); i < arr.length; i++) { arr2[i] = arr[i]; } return arr2; } else { return Array.from(arr); } }

function _classCallCheck(instance, Constructor) { if (!(instance instanceof Constructor)) { throw new TypeError("Cannot call a class as a function"); } }

function _possibleConstructorReturn(self, call) { if (!self) { throw new ReferenceError("this hasn't been initialised - super() hasn't been called"); } return call && (typeof call === "object" || typeof call === "function") ? call : self; }

function _inherits(subClass, superClass) { if (typeof superClass !== "function" && superClass !== null) { throw new TypeError("Super expression must either be null or a function, not " + typeof superClass); } subClass.prototype = Object.create(superClass && superClass.prototype, { constructor: { value: subClass, enumerable: false, writable: true, configurable: true } }); if (superClass) Object.setPrototypeOf ? Object.setPrototypeOf(subClass, superClass) : subClass.__proto__ = superClass; } /**
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                * Graphology Reference Implementation
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                * ====================================
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                *
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                * Reference implementation of the graphology specs.
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                *
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                * Note: Even if the implementation could beneficiate from an abstraction
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                * over the object/map manipulation, it does not for performance reasons.
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                * Another solution would be to split the classes but this would have quite
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                * a cost on the implementation.
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                */


// TODO: finish #.selfLoops (bunch iteration only)

/**
 * Enums.
 */
var TYPES = new _utils.BasicSet(['directed', 'undirected', 'mixed']),
    EMITTER_PROPS = new _utils.BasicSet(['domain', '_events', '_eventsCount', '_maxListeners']);

/**
 * Default options.
 */
var DEFAULTS = {
  allowSelfLoops: true,
  defaultEdgeAttributes: {},
  defaultNodeAttributes: {},
  edgeKeyGenerator: _utils.uuid,
  map: false,
  multi: false,
  onDuplicateEdge: null,
  onDuplicateNode: null,
  type: 'mixed'
};

/**
 * Graph class
 *
 * @constructor
 * @param  {Graph|Array<Array>} [data]    - Hydratation data.
 * @param  {object}             [options] - Options:
 * @param  {boolean}              [allowSelfLoops] - Allow self loops?
 * @param  {string}               [type]           - Type of the graph.
 * @param  {boolean}              [map]            - Allow references as keys?
 * @param  {boolean}              [multi]          - Allow parallel edges?
 *
 * @throws {Error} - Will throw if the arguments are not valid.
 */

var Graph = function (_EventEmitter) {
  _inherits(Graph, _EventEmitter);

  function Graph(data, options) {
    _classCallCheck(this, Graph);

    //-- Solving options
    var _this = _possibleConstructorReturn(this, _EventEmitter.call(this));

    options = (0, _utils.assign)({}, DEFAULTS, options);

    // Enforcing options validity
    if (typeof options.edgeKeyGenerator !== 'function') throw new _errors.InvalidArgumentsGraphError('Graph.constructor: invalid \'edgeKeyGenerator\' option. Expecting a function but got "' + options.map + '".');

    if (typeof options.map !== 'boolean') throw new _errors.InvalidArgumentsGraphError('Graph.constructor: invalid \'map\' option. Expecting a boolean but got "' + options.map + '".');

    if (options.map && typeof Map !== 'function') throw new _errors.InvalidArgumentsGraphError('Graph.constructor: it seems you created a GraphMap instance while your current JavaScript engine does not support ES2015 Map objects.');

    if (typeof options.multi !== 'boolean') throw new _errors.InvalidArgumentsGraphError('Graph.constructor: invalid \'multi\' option. Expecting a boolean but got "' + options.multi + '".');

    if (!TYPES.has(options.type)) throw new _errors.InvalidArgumentsGraphError('Graph.constructor: invalid \'type\' option. Should be one of "mixed", "directed" or "undirected" but got "' + options.type + '".');

    if (typeof options.allowSelfLoops !== 'boolean') throw new _errors.InvalidArgumentsGraphError('Graph.constructor: invalid \'allowSelfLoops\' option. Expecting a boolean but got "' + options.allowSelfLoops + '".');

    if (!(0, _utils.isPlainObject)(options.defaultEdgeAttributes)) throw new _errors.InvalidArgumentsGraphError('Graph.constructor: invalid \'defaultEdgeAttributes\' option. Expecting a plain object but got "' + options.defaultEdgeAttributes + '".');

    if (!(0, _utils.isPlainObject)(options.defaultNodeAttributes)) throw new _errors.InvalidArgumentsGraphError('Graph.constructor: invalid \'defaultNodeAttributes\' option. Expecting a plain object but got "' + options.defaultNodeAttributes + '".');

    if (options.onDuplicateEdge && typeof options.onDuplicateEdge !== 'function') throw new _errors.InvalidArgumentsGraphError('Graph.constructor: invalid \'onDuplicateEdge\' option. Expecting a function but got "' + options.onDuplicateEdge + '".');

    if (options.onDuplicateNode && typeof options.onDuplicateNode !== 'function') throw new _errors.InvalidArgumentsGraphError('Graph.constructor: invalid \'onDuplicateNode\' option. Expecting a function but got "' + options.onDuplicateNode + '".');

    //-- Private properties

    // Counters
    (0, _utils.privateProperty)(_this, '_order', 0);
    (0, _utils.privateProperty)(_this, '_size', 0);

    // Indexes
    (0, _utils.privateProperty)(_this, '_nodes', options.map ? new Map() : {});
    (0, _utils.privateProperty)(_this, '_edges', options.map ? new Map() : {});
    (0, _utils.privateProperty)(_this, '_indices', {
      structure: {
        lazy: options.indices && options.indices.structure && options.indices.structure.lazy || false,
        computed: false
      }
    });

    // Options
    (0, _utils.privateProperty)(_this, '_options', options);

    // Methods
    (0, _utils.privateProperty)(_this, '_addEdge', _this._addEdge);
    (0, _utils.privateProperty)(_this, '_exportEdges', _this._exportEdges);
    (0, _utils.privateProperty)(_this, '_updateIndex', _this._updateIndex);
    (0, _utils.privateProperty)(_this, '_clearEdgeFromIndex', _this._clearEdgeFromIndex);
    (0, _utils.privateProperty)(_this, 'internals', _this.internals);

    //-- Properties readers
    (0, _utils.readOnlyProperty)(_this, 'order', function () {
      return _this._order;
    });
    (0, _utils.readOnlyProperty)(_this, 'size', function () {
      return _this._size;
    });
    (0, _utils.readOnlyProperty)(_this, 'map', function () {
      return _this._options.map;
    });
    (0, _utils.readOnlyProperty)(_this, 'multi', function () {
      return _this._options.multi;
    });
    (0, _utils.readOnlyProperty)(_this, 'type', function () {
      return _this._options.type;
    });
    (0, _utils.readOnlyProperty)(_this, 'allowSelfLoops', function () {
      return _this._options.allowSelfLoops;
    });

    //-- Precomputing indexes?
    for (var name in _this._indices) {
      var index = _this._indices[name];

      if (!index.lazy) index.computed = true;
    }

    //-- Hydratation
    if (data) _this.import(data);
    return _this;
  }

  /**---------------------------------------------------------------------------
   * Read
   **---------------------------------------------------------------------------
   */

  /**
   * Method returning whether the given node is found in the graph.
   *
   * @param  {any}     node - The node.
   * @return {boolean}
   */


  Graph.prototype.hasNode = function hasNode(node) {
    var nodeInGraph = false;

    if (this.map) nodeInGraph = this._nodes.has(node);else nodeInGraph = node in this._nodes;

    return nodeInGraph;
  };

  /**
   * Internal method returning a matching directed edge or undefined if no
   * matching edge was found.
   *
   * @param  {any}     source - The edge's source.
   * @param  {any}     target - The edge's target.
   * @return {any|undefined}
   */


  Graph.prototype.getDirectedEdge = function getDirectedEdge(source, target) {

    // We need to compute the 'structure' index for this
    this.computeIndex('structure');

    // If the node source or the target is not in the graph we break
    if (!this.hasNode(source) || !this.hasNode(target)) return;

    // Is there a directed edge pointing towards target?
    var nodeData = this.map ? this._nodes.get(source) : this._nodes[source];

    var register = nodeData.out;

    if (!register) return;

    var edges = this.map ? register.get(target) : register[target];

    if (!edges) return;

    if (!edges.size) return;

    return this.map ? edges.values().next().value : edges.first();
  };

  /**
   * Internal method returning a matching undirected edge or undefined if no
   * matching edge was found.
   *
   * @param  {any}     source - The edge's source.
   * @param  {any}     target - The edge's target.
   * @return {any|undefined}
   */


  Graph.prototype.getUndirectedEdge = function getUndirectedEdge(source, target) {

    // We need to compute the 'structure' index for this
    this.computeIndex('structure');

    // If the node source or the target is not in the graph we break
    if (!this.hasNode(source) || !this.hasNode(target)) return;

    // Is there a directed edge pointing towards target?
    var nodeData = this.map ? this._nodes.get(source) : this._nodes[source];

    var register = nodeData.undirectedOut,
        edges = void 0;

    if (register) edges = this.map ? register.get(target) : register[target];

    register = nodeData.undirectedIn;

    if (!edges && register) edges = this.map ? register.get(target) : register[target];

    if (!edges) return;

    if (!edges.size) return;

    return this.map ? edges.values().next().value : edges.first();
  };

  /**
   * Method returning a matching edge (note that it will return the first
   * matching edge, starting with directed one then undirected), or undefined
   * if no matching edge was found.
   *
   * @param  {any}     source - The edge's source.
   * @param  {any}     target - The edge's target.
   * @return {any|undefined}
   */


  Graph.prototype.getEdge = function getEdge(source, target) {
    var edge = void 0;

    // First we try to find a directed edge
    if (this.type === 'mixed' || this.type === 'directed') edge = this.getDirectedEdge(source, target);

    if (edge) return edge;

    // Then we try to find an undirected edge
    if (this.type === 'mixed' || this.type === 'undirected') edge = this.getUndirectedEdge(source, target);

    return edge;
  };

  /**
   * Method returning whether the given directed edge is found in the graph.
   *
   * Arity 1:
   * @param  {any}     edge - The edge's key.
   *
   * Arity 2:
   * @param  {any}     source - The edge's source.
   * @param  {any}     target - The edge's target.
   *
   * @return {boolean}
   *
   * @throws {Error} - Will throw if the arguments are invalid.
   */


  Graph.prototype.hasDirectedEdge = function hasDirectedEdge(source, target) {
    if (arguments.length === 1) {
      var edge = source;

      return (this.map ? this._edges.has(edge) : edge in this._edges) && this.directed(edge);
    } else if (arguments.length === 2) {

      // We need to compute the 'structure' index for this
      this.computeIndex('structure');

      // If the node source or the target is not in the graph we break
      if (!this.hasNode(source) || !this.hasNode(target)) return false;

      // Is there a directed edge pointing towards target?
      var nodeData = this.map ? this._nodes.get(source) : this._nodes[source];

      var register = nodeData.out;

      if (!register) return false;

      var edges = this.map ? register.get(target) : register[target];

      if (!edges) return false;

      return !!edges.size;
    }

    throw new _errors.InvalidArgumentsGraphError('Graph.hasDirectedEdge: invalid arity (' + arguments.length + ', instead of 1 or 2). You can either ask for an edge id or for the existence of an edge between a source & a target.');
  };

  /**
   * Method returning whether the given undirected edge is found in the graph.
   *
   * Arity 1:
   * @param  {any}     edge - The edge's key.
   *
   * Arity 2:
   * @param  {any}     source - The edge's source.
   * @param  {any}     target - The edge's target.
   *
   * @return {boolean}
   *
   * @throws {Error} - Will throw if the arguments are invalid.
   */


  Graph.prototype.hasUndirectedEdge = function hasUndirectedEdge(source, target) {
    if (arguments.length === 1) {
      var edge = source;

      return (this.map ? this._edges.has(edge) : edge in this._edges) && this.undirected(edge);
    } else if (arguments.length === 2) {

      // We need to compute the 'structure' index for this
      this.computeIndex('structure');

      // If the node source or the target is not in the graph we break
      if (!this.hasNode(source) || !this.hasNode(target)) return false;

      // Is there a directed edge pointing towards target?
      var nodeData = this.map ? this._nodes.get(source) : this._nodes[source];

      var register = nodeData.undirectedOut,
          edges = void 0;

      if (register) edges = this.map ? register.get(target) : register[target];

      register = nodeData.undirectedIn;

      if (!edges && register) edges = this.map ? register.get(target) : register[target];

      if (!edges) return false;

      return !!edges.size;
    }

    throw new _errors.InvalidArgumentsGraphError('Graph.hasDirectedEdge: invalid arity (' + arguments.length + ', instead of 1 or 2). You can either ask for an edge id or for the existence of an edge between a source & a target.');
  };

  /**
   * Method returning whether the given edge is found in the graph.
   *
   * Arity 1:
   * @param  {any}     edge - The edge's key.
   *
   * Arity 2:
   * @param  {any}     source - The edge's source.
   * @param  {any}     target - The edge's target.
   *
   * @return {boolean}
   *
   * @throws {Error} - Will throw if the arguments are invalid.
   */


  Graph.prototype.hasEdge = function hasEdge(source, target) {

    if (arguments.length === 1) {
      var edge = source;

      return this.map ? this._edges.has(edge) : edge in this._edges;
    } else if (arguments.length === 2) {
      return this.hasDirectedEdge(source, target) || this.hasUndirectedEdge(source, target);
    }

    throw new _errors.InvalidArgumentsGraphError('Graph.hasEdge: invalid arity (' + arguments.length + ', instead of 1 or 2). You can either ask for an edge id or for the existence of an edge between a source & a target.');
  };

  /**
   * Method returning the given node's in degree.
   *
   * @param  {any}     node      - The node's key.
   * @param  {boolean} allowSelfLoops - Count self-loops?
   * @return {number}            - The node's in degree.
   *
   * @throws {Error} - Will throw if the selfLoops arg is not boolean.
   * @throws {Error} - Will throw if the node isn't in the graph.
   */


  Graph.prototype.inDegree = function inDegree(node) {
    var selfLoops = arguments.length <= 1 || arguments[1] === undefined ? true : arguments[1];

    if (typeof selfLoops !== 'boolean') throw new _errors.InvalidArgumentsGraphError('Graph.inDegree: Expecting a boolean but got "' + selfLoops + '" for the second parameter (allowing self-loops to be counted).');

    if (!this.hasNode(node)) throw new _errors.NotFoundGraphError('Graph.inDegree: could not find the "' + node + '" node in the graph.');

    var data = this.map ? this._nodes.get(node) : this._nodes[node];

    return data.inDegree + (selfLoops ? data.selfLoops : 0);
  };

  /**
   * Method returning the given node's out degree.
   *
   * @param  {any}     node      - The node's key.
   * @param  {boolean} selfLoops - Count self-loops?
   * @return {number}            - The node's out degree.
   *
   * @throws {Error} - Will throw if the selfLoops arg is not boolean.
   * @throws {Error} - Will throw if the node isn't in the graph.
   */


  Graph.prototype.outDegree = function outDegree(node) {
    var selfLoops = arguments.length <= 1 || arguments[1] === undefined ? true : arguments[1];

    if (typeof selfLoops !== 'boolean') throw new _errors.InvalidArgumentsGraphError('Graph.outDegree: Expecting a boolean but got "' + selfLoops + '" for the second parameter (allowing self-loops to be counted).');

    if (!this.hasNode(node)) throw new _errors.NotFoundGraphError('Graph.outDegree: could not find the "' + node + '" node in the graph.');

    var data = this.map ? this._nodes.get(node) : this._nodes[node];

    return data.outDegree + (selfLoops ? data.selfLoops : 0);
  };

  /**
   * Method returning the given node's directed degree.
   *
   * @param  {any}     node      - The node's key.
   * @param  {boolean} selfLoops - Count self-loops?
   * @return {number}            - The node's directed degree.
   *
   * @throws {Error} - Will throw if the selfLoops arg is not boolean.
   * @throws {Error} - Will throw if the node isn't in the graph.
   */


  Graph.prototype.directedDegree = function directedDegree(node) {
    var selfLoops = arguments.length <= 1 || arguments[1] === undefined ? true : arguments[1];

    if (typeof selfLoops !== 'boolean') throw new _errors.InvalidArgumentsGraphError('Graph.directedDegree: Expecting a boolean but got "' + selfLoops + '" for the second parameter (allowing self-loops to be counted).');

    if (!this.hasNode(node)) throw new _errors.NotFoundGraphError('Graph.directedDegree: could not find the "' + node + '" node in the graph.');

    var data = this.map ? this._nodes.get(node) : this._nodes[node];

    return data.outDegree + data.inDegree + (selfLoops ? data.selfLoops : 0);
  };

  /**
   * Method returning the given node's undirected degree.
   *
   * @param  {any}     node      - The node's key.
   * @param  {boolean} selfLoops - Count self-loops?
   * @return {number}            - The node's undirected degree.
   *
   * @throws {Error} - Will throw if the selfLoops arg is not boolean.
   * @throws {Error} - Will throw if the node isn't in the graph.
   */


  Graph.prototype.undirectedDegree = function undirectedDegree(node) {
    var selfLoops = arguments.length <= 1 || arguments[1] === undefined ? true : arguments[1];

    if (typeof selfLoops !== 'boolean') throw new _errors.InvalidArgumentsGraphError('Graph.undirectedDegree: Expecting a boolean but got "' + selfLoops + '" for the second parameter (allowing self-loops to be counted).');

    if (!this.hasNode(node)) throw new _errors.NotFoundGraphError('Graph.undirectedDegree: could not find the "' + node + '" node in the graph.');

    var data = this.map ? this._nodes.get(node) : this._nodes[node];

    return data.undirectedDegree + (selfLoops ? data.selfLoops : 0);
  };

  /**
   * Method returning the given node's degree.
   *
   * @param  {any}     node      - The node's key.
   * @param  {boolean} selfLoops - Count self-loops?
   * @return {number}            - The node's degree.
   *
   * @throws {Error} - Will throw if the selfLoops arg is not boolean.
   * @throws {Error} - Will throw if the node isn't in the graph.
   */


  Graph.prototype.degree = function degree(node) {
    var selfLoops = arguments.length <= 1 || arguments[1] === undefined ? true : arguments[1];

    if (typeof selfLoops !== 'boolean') throw new _errors.InvalidArgumentsGraphError('Graph.degree: Expecting a boolean but got "' + selfLoops + '" for the second parameter (allowing self-loops to be counted).');

    if (!this.hasNode(node)) throw new _errors.NotFoundGraphError('Graph.degree: could not find the "' + node + '" node in the graph.');

    var data = this.map ? this._nodes.get(node) : this._nodes[node];

    return data.outDegree + data.inDegree + data.undirectedDegree + (selfLoops ? data.selfLoops : 0);
  };

  /**
   * Method returning the given edge's source.
   *
   * @param  {any} edge - The edge's key.
   * @return {any}      - The edge's source.
   *
   * @throws {Error} - Will throw if the edge isn't in the graph.
   */


  Graph.prototype.source = function source(edge) {
    if (!this.hasEdge(edge)) throw new _errors.NotFoundGraphError('Graph.source: could not find the "' + edge + '" edge in the graph.');

    var source = this.map ? this._edges.get(edge).source : this._edges[edge].source;

    return source;
  };

  /**
   * Method returning the given edge's target.
   *
   * @param  {any} edge - The edge's key.
   * @return {any}      - The edge's target.
   *
   * @throws {Error} - Will throw if the edge isn't in the graph.
   */


  Graph.prototype.target = function target(edge) {
    if (!this.hasEdge(edge)) throw new _errors.NotFoundGraphError('Graph.target: could not find the "' + edge + '" edge in the graph.');

    var target = this.map ? this._edges.get(edge).target : this._edges[edge].target;

    return target;
  };

  /**
   * Method returning the given edge's extremities.
   *
   * @param  {any}   edge - The edge's key.
   * @return {array}      - The edge's extremities.
   *
   * @throws {Error} - Will throw if the edge isn't in the graph.
   */


  Graph.prototype.extremities = function extremities(edge) {
    if (!this.hasEdge(edge)) throw new _errors.NotFoundGraphError('Graph.extremities: could not find the "' + edge + '" edge in the graph.');

    return [this.source(edge), this.target(edge)];
  };

  /**
   * Given a node & an edge, returns the other extremity of the edge.
   *
   * @param  {any}   node - The node's key.
   * @param  {any}   edge - The edge's key.
   * @return {any}        - The related node.
   *
   * @throws {Error} - Will throw if either the node or the edge isn't in the graph.
   */


  Graph.prototype.relatedNode = function relatedNode(node, edge) {
    if (!this.hasNode(node)) throw new _errors.NotFoundGraphError('Graph.relatedNode: could not find the "' + node + '" node in the graph.');

    if (!this.hasEdge(edge)) throw new _errors.NotFoundGraphError('Graph.relatedNode: could not find the "' + edge + '" edge in the graph.');

    var _extremities = this.extremities(edge);

    var _extremities2 = _slicedToArray(_extremities, 2);

    var node1 = _extremities2[0];
    var node2 = _extremities2[1];


    return node === node1 ? node2 : node1;
  };

  /**
   * Method returning whether the given edge is undirected.
   *
   * @param  {any}     edge - The edge's key.
   * @return {boolean}
   *
   * @throws {Error} - Will throw if the edge isn't in the graph.
   */


  Graph.prototype.undirected = function undirected(edge) {
    if (!this.hasEdge(edge)) throw new _errors.NotFoundGraphError('Graph.undirected: could not find the "' + edge + '" edge in the graph.');

    var undirected = this.map ? this._edges.get(edge).undirected : this._edges[edge].undirected;

    return !!undirected;
  };

  /**
   * Method returning whether the given edge is directed.
   *
   * @param  {any}     edge - The edge's key.
   * @return {boolean}
   *
   * @throws {Error} - Will throw if the edge isn't in the graph.
   */


  Graph.prototype.directed = function directed(edge) {
    if (!this.hasEdge(edge)) throw new _errors.NotFoundGraphError('Graph.directed: could not find the "' + edge + '" edge in the graph.');

    var undirected = this.map ? this._edges.get(edge).undirected : this._edges[edge].undirected;

    return !undirected;
  };

  /**
   * Method returning whether the given edge is a self loop.
   *
   * @param  {any}     edge - The edge's key.
   * @return {boolean}
   *
   * @throws {Error} - Will throw if the edge isn't in the graph.
   */


  Graph.prototype.selfLoop = function selfLoop(edge) {
    if (!this.hasEdge(edge)) throw new _errors.NotFoundGraphError('Graph.selfLoop: could not find the "' + edge + '" edge in the graph.');

    var data = this.map ? this._edges.get(edge) : this._edges[edge];

    return data.source === data.target;
  };

  /**---------------------------------------------------------------------------
   * Mutation
   **---------------------------------------------------------------------------
   */

  /**
   * Method used to add a node to the graph.
   *
   * @param  {any}    node         - The node.
   * @param  {object} [attributes] - Optional attributes.
   * @return {any}                 - The node.
   *
   * @throws {Error} - Will throw if the given node already exist.
   * @throws {Error} - Will throw if the given attributes are not an object.
   */


  Graph.prototype.addNode = function addNode(node, attributes) {
    if (arguments.length > 1 && !(0, _utils.isPlainObject)(attributes)) throw new _errors.InvalidArgumentsGraphError('Graph.addNode: invalid attributes. Expecting an object but got "' + attributes + '"');

    // Protecting the attributes
    attributes = (0, _utils.assign)({}, this._options.defaultNodeAttributes, attributes);

    if (this.hasNode(node)) {

      // Triggering duplicate callback
      if (typeof this._options.onDuplicateNode === 'function') {
        this._options.onDuplicateNode(this, { key: node, attributes: attributes });

        return node;
      } else {
        throw new _errors.UsageGraphError('Graph.addNode: the "' + node + '" node already exist in the graph. You might want to check out the \'onDuplicateNode\' option.');
      }
    }

    var data = {
      attributes: attributes
    };

    if (this.allowSelfLoops) data.selfLoops = 0;

    if (this.type === 'mixed' || this.type === 'directed') {
      data.inDegree = 0;
      data.outDegree = 0;
    }

    if (this.type === 'mixed' || this.type === 'undirected') {
      data.undirectedDegree = 0;
    }

    // Adding the node to internal register
    if (this.map) this._nodes.set(node, data);else this._nodes[node] = data;

    // Incrementing order
    this._order++;

    // Emitting
    this.emit('nodeAdded', {
      key: node,
      attributes: attributes
    });

    return node;
  };

  /**
   * Method used to add a nodes from a bunch.
   *
   * @param  {bunch}  bunch - The node.
   * @return {Graph}        - Returns itself for chaining.
   *
   * @throws {Error} - Will throw if the given bunch is not valid.
   */


  Graph.prototype.addNodesFrom = function addNodesFrom(bunch) {
    var _this2 = this;

    if (!(0, _utils.isBunch)(bunch)) throw new _errors.InvalidArgumentsGraphError('Graph.addNodesFrom: invalid bunch provided ("' + bunch + '").');

    (0, _utils.overBunch)(bunch, function (error, node, attributes) {
      _this2.addNode(node, attributes);
    });

    return this;
  };

  /**
   * Internal method used to add an arbitrary edge to the graph.
   *
   * @param  {string}  name         - Name of the child method for errors.
   * @param  {boolean} undirected   - Whether the edge is undirected.
   * @param  {any}     edge         - The edge's key.
   * @param  {any}     source       - The source node.
   * @param  {any}     target       - The target node.
   * @param  {object}  [attributes] - Optional attributes.
   * @return {any}                  - The edge.
   *
   * @throws {Error} - Will throw if the graph is undirected.
   * @throws {Error} - Will throw if the given attributes are not an object.
   * @throws {Error} - Will throw if source or target doesn't exist.
   * @throws {Error} - Will throw if the edge already exist.
   */


  Graph.prototype._addEdge = function _addEdge(name, undirected, edge, source, target, attributes) {

    if (!undirected && this.type === 'undirected') throw new _errors.UsageGraphError('Graph.' + name + ': you cannot add a directed edge to an undirected graph. Use the #.addEdge or #.addUndirectedEdge instead.');

    if (undirected && this.type === 'directed') throw new _errors.UsageGraphError('Graph.' + name + ': you cannot add an undirected edge to a directed graph. Use the #.addEdge or #.addDirectedEdge instead.');

    if (attributes && !(0, _utils.isPlainObject)(attributes)) throw new _errors.InvalidArgumentsGraphError('Graph.' + name + ': invalid attributes. Expecting an object but got "' + attributes + '"');

    if (!this.hasNode(source)) throw new _errors.NotFoundGraphError('Graph.' + name + ': source node "' + source + '" not found.');

    if (!this.hasNode(target)) throw new _errors.NotFoundGraphError('Graph.' + name + ': target node "' + target + '" not found.');

    if (!this.allowSelfLoops && source === target) throw new _errors.UsageGraphError('Graph.' + name + ': source & target are the same, thus creating a loop explicitly forbidden by this graph \'allowSelfLoops\' option set to false.');

    var canHandleDuplicate = typeof this._options.onDuplicateEdge === 'function';
    var mustHandleDuplicate = false;

    if (this.hasEdge(edge)) {
      if (!canHandleDuplicate) throw new _errors.UsageGraphError('Graph.' + name + ': the "' + edge + '" edge already exists in the graph.');else mustHandleDuplicate = true;
    }

    if (!this.multi && (undirected ? this.hasUndirectedEdge(source, target) : this.hasDirectedEdge(source, target))) {
      if (!canHandleDuplicate) throw new _errors.UsageGraphError('Graph.' + name + ': an edge linking "' + source + '" to "' + target + '" already exists. If you really want to add multiple edges linking those nodes, you should create a multi graph by using the \'multi\' option. The \'onDuplicateEdge\' option might also interest you.');else mustHandleDuplicate = true;
    }

    // Protecting the attributes
    attributes = (0, _utils.assign)({}, this._options.defaultEdgeAttributes, attributes);

    // Handling duplicates
    if (mustHandleDuplicate) {

      // TODO: decide whether to stock that id was generated and what to pass
      // here
      this._options.onDuplicateEdge(this, {
        key: edge,
        attributes: attributes,
        source: source,
        target: target,
        undirected: undirected
      });

      return edge;
    }

    // Storing some data
    var data = {
      attributes: attributes,
      source: source,
      target: target
    };

    // NOTE: only adding the 'undirected' key if needed
    if (undirected) data.undirected = true;

    if (this.map) this._edges.set(edge, data);else this._edges[edge] = data;

    // Incrementing size
    this._size++;

    // Incrementing node counters
    var sourceData = this.map ? this._nodes.get(source) : this._nodes[source],
        targetData = this.map ? this._nodes.get(target) : this._nodes[target];

    if (source === target) {
      sourceData.selfLoops++;
    } else {
      if (undirected) {
        sourceData.undirectedDegree++;
        targetData.undirectedDegree++;
      } else {
        sourceData.outDegree++;
        targetData.inDegree++;
      }
    }

    // Updating relevant indexes
    this._updateIndex('structure', edge, data);

    // Emitting
    this.emit('edgeAdded', {
      key: edge,
      source: source,
      target: target,
      attributes: attributes,
      undirected: undirected
    });

    return edge;
  };

  /**
   * Method used to add an edge of the type of the graph or directed if the
   * graph is mixed using the given key.
   *
   * @param  {any}    edge         - The edge's key.
   * @param  {any}    source       - The source node.
   * @param  {any}    target       - The target node.
   * @param  {object} [attributes] - Optional attributes.
   * @return {any}                 - The edge.
   */


  Graph.prototype.addEdgeWithKey = function addEdgeWithKey(edge, source, target, attributes) {
    return this._addEdge('addEdgeWithKey', this.type === 'undirected', edge, source, target, attributes);
  };

  /**
   * Method used to add a directed edge to the graph using the given key.
   *
   * @param  {any}    edge         - The edge's key.
   * @param  {any}    source       - The source node.
   * @param  {any}    target       - The target node.
   * @param  {object} [attributes] - Optional attributes.
   * @return {any}                 - The edge.
   */


  Graph.prototype.addDirectedEdgeWithKey = function addDirectedEdgeWithKey(edge, source, target, attributes) {
    return this._addEdge('addDirectedEdgeWithKey', false, edge, source, target, attributes);
  };

  /**
   * Method used to add an undirected edge to the graph using the given key.
   *
   * @param  {any}    edge         - The edge's key.
   * @param  {any}    source       - The source node.
   * @param  {any}    target       - The target node.
   * @param  {object} [attributes] - Optional attributes.
   * @return {any}                 - The edge.
   */


  Graph.prototype.addUndirectedEdgeWithKey = function addUndirectedEdgeWithKey(edge, source, target, attributes) {
    return this._addEdge('addUndirectedEdgeWithKey', true, edge, source, target, attributes);
  };

  /**
   * Method used to add an edge of the type of the graph or directed if the
   * graph is mixed. An id will automatically be created for it using the
   * 'edgeKeyGenerator' option.
   *
   * @param  {any}    source       - The source node.
   * @param  {any}    target       - The target node.
   * @param  {object} [attributes] - Optional attributes.
   * @return {any}                 - The edge.
   */


  Graph.prototype.addEdge = function addEdge(source, target, attributes) {
    var edge = this._options.edgeKeyGenerator(this.type === 'undirected', source, target, attributes);

    return this._addEdge('addEdge', this.type === 'undirected', edge, source, target, attributes);
  };

  /**
   * Method used to add a directed edge to the graph. An id will automatically
   * be created for it using the 'edgeKeyGenerator' option.
   *
   * @param  {any}    source       - The source node.
   * @param  {any}    target       - The target node.
   * @param  {object} [attributes] - Optional attributes.
   * @return {any}                 - The edge.
   */


  Graph.prototype.addDirectedEdge = function addDirectedEdge(source, target, attributes) {

    // Generating an id
    var edge = this._options.edgeKeyGenerator(false, source, target, attributes);

    return this._addEdge('addDirectedEdge', false, edge, source, target, attributes);
  };

  /**
   * Method used to add an undirected edge to the graph. An id will automatically
   * be created for it using the 'edgeKeyGenerator' option.
   *
   * @param  {any}    source       - The source node.
   * @param  {any}    target       - The target node.
   * @param  {object} [attributes] - Optional attributes.
   * @return {any}                 - The edge.
   */


  Graph.prototype.addUndirectedEdge = function addUndirectedEdge(source, target, attributes) {

    // Generating an id
    var edge = this._options.edgeKeyGenerator(false, source, target, attributes);

    return this._addEdge('addUndirectedEdge', true, edge, source, target, attributes);
  };

  /**
   * Method used to drop a single node & all its attached edges from the graph.
   *
   * @param  {any}    node - The node.
   * @return {Graph}
   *
   * @throws {Error} - Will throw if the node doesn't exist.
   */


  Graph.prototype.dropNode = function dropNode(node) {
    if (!this.hasNode(node)) throw new _errors.NotFoundGraphError('Graph.dropNode: could not find the "' + node + '" node in the graph.');

    // Removing attached edges
    var edges = this.edges(node);

    // NOTE: we could go faster here
    for (var i = 0, l = edges.length; i < l; i++) {
      this.dropEdge(edges[i]);
    }var data = this.map ? this._nodes.get(node) : this._nodes[node];

    // Dropping the node from the register
    if (this.map) this._nodes.delete(node);else delete this._nodes[node];

    // Decrementing order
    this._order--;

    // Emitting
    this.emit('nodeDropped', {
      key: node,
      attributes: data.attributes
    });
  };

  /**
   * Method used to drop a single edge from the graph.
   *
   * Arity 1:
   * @param  {any}    edge - The edge.
   *
   * Arity 2:
   * @param  {any}    source - Source node.
   * @param  {any}    target - Target node.
   *
   * @return {Graph}
   *
   * @throws {Error} - Will throw if the edge doesn't exist.
   */


  Graph.prototype.dropEdge = function dropEdge(edge) {
    if (arguments.length > 1) {
      var _source = arguments[0],
          _target = arguments[1];

      if (!this.hasNode(_source)) throw new _errors.NotFoundGraphError('Graph.dropEdge: could not find the "' + _source + '" source node in the graph.');

      if (!this.hasNode(_target)) throw new _errors.NotFoundGraphError('Graph.dropEdge: could not find the "' + _target + '" target node in the graph.');

      if (!this.hasEdge(_source, _target)) throw new _errors.NotFoundGraphError('Graph.dropEdge: could not find the "' + _source + '" -> "' + _target + '" edge in the graph.');

      edge = this.getEdge(_source, _target);
    } else {
      if (!this.hasEdge(edge)) throw new _errors.NotFoundGraphError('Graph.dropEdge: could not find the "' + edge + '" edge in the graph.');
    }

    var data = this.map ? this._edges.get(edge) : this._edges[edge];

    // Dropping the edge from the register
    if (this.map) this._edges.delete(edge);else delete this._edges[edge];

    // Decrementing size
    this._size--;

    // Updating related degrees
    var source = data.source;
    var target = data.target;
    var attributes = data.attributes;
    var _data$undirected = data.undirected;
    var undirected = _data$undirected === undefined ? false : _data$undirected;


    var sourceData = this.map ? this._nodes.get(source) : this._nodes[source],
        targetData = this.map ? this._nodes.get(target) : this._nodes[target];

    if (source === target) {
      sourceData.selfLoops--;
    } else {
      if (undirected) {
        sourceData.undirectedDegree--;
        targetData.undirectedDegree--;
      } else {
        sourceData.outDegree--;
        targetData.inDegree--;
      }
    }

    // Clearing index
    this._clearEdgeFromIndex('structure', edge, data);

    // Emitting
    this.emit('edgeDropped', {
      key: edge,
      attributes: attributes,
      source: source,
      target: target,
      undirected: undirected
    });

    return this;
  };

  /**
   * Method used to drop a bunch of nodes or every node from the graph.
   *
   * @param  {bunch} nodes - Bunch of nodes.
   * @return {Graph}
   *
   * @throws {Error} - Will throw if an invalid bunch is provided.
   * @throws {Error} - Will throw if any of the nodes doesn't exist.
   */


  Graph.prototype.dropNodes = function dropNodes(nodes) {
    var _this3 = this;

    if (!arguments.length) return this.clear();

    if (!(0, _utils.isBunch)(nodes)) throw new _errors.InvalidArgumentsGraphError('Graph.dropNodes: invalid bunch.');

    (0, _utils.overBunch)(nodes, function (error, node) {
      _this3.dropNode(node);
    });

    return this;
  };

  /**
   * Method used to drop a bunch of edges or every edges from the graph.
   *
   * Arity 1:
   * @param  {bunch} edges - Bunch of edges.
   *
   * Arity 2:
   * @param  {any}    source - Source node.
   * @param  {any}    target - Target node.
   *
   * @return {Graph}
   *
   * @throws {Error} - Will throw if an invalid bunch is provided.
   * @throws {Error} - Will throw if any of the edges doesn't exist.
   */


  Graph.prototype.dropEdges = function dropEdges(edges) {
    var _this4 = this;

    if (!arguments.length) {

      // Dropping every edge from the graph
      this._edges = this.map ? new Map() : {};
      this._size = 0;

      // Without edges, we've got no 'structure'
      this.clearIndex('structure');

      var index = this._indices.structure;

      if (!index.lazy) index.computed = true;

      return this;
    }

    if (arguments.length === 2) {
      var source = arguments[0],
          target = arguments[1];

      edges = this.edges(source, target);
    }

    if (!(0, _utils.isBunch)(edges)) throw new _errors.InvalidArgumentsGraphError('Graph.dropEdges: invalid bunch.');

    (0, _utils.overBunch)(edges, function (error, edge) {
      _this4.dropEdge(edge);
    });

    return this;
  };

  /**
   * Method used to remove every edge & every node from the graph.
   *
   * @return {Graph}
   */


  Graph.prototype.clear = function clear() {

    // Dropping edges
    this._edges = this.map ? new Map() : {};

    // Dropping nodes
    this._nodes = this.map ? new Map() : {};

    // Resetting counters
    this._order = 0;
    this._size = 0;

    for (var name in this._indices) {
      var index = this._indices[name];

      if (index.lazy) index.computed = false;
    }

    // Emitting
    this.emit('cleared');
  };

  /**---------------------------------------------------------------------------
   * Iteration-related methods
   **---------------------------------------------------------------------------
   */

  /**
   * Method returning the list of the graph's nodes.
   *
   * @return {array} - The nodes.
   */


  Graph.prototype.nodes = function nodes() {

    if (this.map) return [].concat(_toConsumableArray(this._nodes.keys()));

    return Object.keys(this._nodes);
  };

  /**---------------------------------------------------------------------------
   * Serialization
   **---------------------------------------------------------------------------
   */

  /**
   * Method exporting the target node.
   *
   * @param  {any}   node - Target node.
   * @return {array}      - The serialized node.
   *
   * @throws {Error} - Will throw if the node is not found.
   */


  Graph.prototype.exportNode = function exportNode(node) {
    if (!this.hasNode(node)) throw new _errors.NotFoundGraphError('Graph.exportNode: could not find the "' + node + '" node in the graph.');

    var data = this.map ? this._nodes.get(node) : this._nodes[node];

    return (0, _serialization.serializeNode)(node, data);
  };

  /**
   * Method exporting the target edge.
   *
   * @param  {any}   edge - Target edge.
   * @return {array}      - The serialized edge.
   *
   * @throws {Error} - Will throw if the edge is not found.
   */


  Graph.prototype.exportEdge = function exportEdge(edge) {
    if (!this.hasEdge(edge)) throw new _errors.NotFoundGraphError('Graph.exportEdge: could not find the "' + edge + '" edge in the graph.');

    var data = this.map ? this._edges.get(edge) : this._edges[edge];

    return (0, _serialization.serializeEdge)(edge, data);
  };

  /**
   * Method exporting every nodes or the bunch ones.
   *
   * @param  {mixed}   [bunch] - Target nodes.
   * @return {array[]}         - The serialized nodes.
   *
   * @throws {Error} - Will throw if any of the nodes is not found.
   */


  Graph.prototype.exportNodes = function exportNodes(bunch) {
    var _this5 = this;

    var nodes = [];

    if (!arguments.length) {

      // Exporting every node
      nodes = this.nodes();
    } else {

      // Exporting the bunch
      if (!(0, _utils.isBunch)(bunch)) throw new _errors.InvalidArgumentsGraphError('Graph.exportNodes: invalid bunch.');

      (0, _utils.overBunch)(bunch, function (error, node) {
        if (!_this5.hasNode(node)) throw new _errors.NotFoundGraphError('Graph.exportNodes: could not find the "' + node + '" node from the bunch in the graph.');
        nodes.push(node);
      });
    }

    var serializedNodes = new Array(nodes.length);

    for (var i = 0, l = nodes.length; i < l; i++) {
      serializedNodes[i] = this.exportNode(nodes[i]);
    }return serializedNodes;
  };

  /**
   * Internal method abstracting edges export.
   *
   * @param  {string}   name      - Child method name.
   * @param  {function} predicate - Predicate to filter the bunch's edges.
   * @param  {mixed}    [bunch]   - Target edges.
   * @return {array[]}            - The serialized edges.
   *
   * @throws {Error} - Will throw if any of the edges is not found.
   */


  Graph.prototype._exportEdges = function _exportEdges(name, predicate, bunch) {
    var _this6 = this;

    var edges = [];

    if (!bunch) {

      // Exporting every edges of the given type
      if (name === 'exportEdges') edges = this.edges();else if (name === 'exportDirectedEdges') edges = this.directedEdges();else edges = this.undirectedEdges();
    } else {

      // Exporting the bunch
      if (!(0, _utils.isBunch)(bunch)) throw new _errors.InvalidArgumentsGraphError('Graph.' + name + ': invalid bunch.');

      (0, _utils.overBunch)(bunch, function (error, edge) {
        if (!_this6.hasEdge(edge)) throw new _errors.NotFoundGraphError('Graph.' + name + ': could not find the "' + edge + '" edge from the bunch in the graph.');

        if (!predicate || predicate(edge)) edges.push(edge);
      });
    }

    var serializedEdges = new Array(edges.length);

    for (var i = 0, l = edges.length; i < l; i++) {
      serializedEdges[i] = this.exportEdge(edges[i]);
    }return serializedEdges;
  };

  /**
   * Method exporting every edges or the bunch ones.
   *
   * @param  {mixed}   [bunch] - Target edges.
   * @return {array[]}         - The serialized edges.
   *
   * @throws {Error} - Will throw if any of the edges is not found.
   */


  Graph.prototype.exportEdges = function exportEdges(bunch) {
    return this._exportEdges('exportEdges', null, bunch);
  };

  /**
   * Method exporting every directed edges or the bunch ones which are directed.
   *
   * @param  {mixed}   [bunch] - Target edges.
   * @return {array[]}         - The serialized edges.
   *
   * @throws {Error} - Will throw if any of the edges is not found.
   */


  Graph.prototype.exportDirectedEdges = function exportDirectedEdges(bunch) {
    var _this7 = this;

    return this._exportEdges('exportDirectedEdges', function (edge) {
      return _this7.directed(edge);
    }, bunch);
  };

  /**
   * Method exporting every unddirected edges or the bunch ones which are
   * undirected
   *
   * @param  {mixed}   [bunch] - Target edges.
   * @return {array[]}         - The serialized edges.
   *
   * @throws {Error} - Will throw if any of the edges is not found.
   */


  Graph.prototype.exportUndirectedEdges = function exportUndirectedEdges(bunch) {
    var _this8 = this;

    return this._exportEdges('exportUndirectedEdges', function (edge) {
      return _this8.undirected(edge);
    }, bunch);
  };

  /**
   * Method used to export the whole graph.
   *
   * @return {object} - The serialized graph.
   */


  Graph.prototype.export = function _export() {
    return {
      nodes: this.exportNodes(),
      edges: this.exportEdges()
    };
  };

  /**
   * Method used to import a serialized node.
   *
   * @param  {object} data - The serialized node.
   * @return {Graph}       - Returns itself for chaining.
   */


  Graph.prototype.importNode = function importNode(data) {

    // Validating
    var _validateSerializedNo = (0, _serialization.validateSerializedNode)(data);

    var valid = _validateSerializedNo.valid;
    var reason = _validateSerializedNo.reason;


    if (!valid) {
      if (reason === 'not-object') throw new _errors.InvalidArgumentsGraphError('Graph.importNode: invalid serialized node. A serialized node should be a plain object with at least a "key" property.');
      if (reason === 'no-key') throw new _errors.InvalidArgumentsGraphError('Graph.importNode: no key provided.');
      if (reason === 'invalid-attributes') throw new _errors.InvalidArgumentsGraphError('Graph.importNode: invalid attributes. Attributes should be a plain object, null or omitted.');
    }

    // Adding the node
    var key = data.key;
    var _data$attributes = data.attributes;
    var attributes = _data$attributes === undefined ? {} : _data$attributes;


    this.addNode(key, attributes);

    return this;
  };

  /**
   * Method used to import a serialized edge.
   *
   * @param  {object} data - The serialized edge.
   * @return {Graph}       - Returns itself for chaining.
   */


  Graph.prototype.importEdge = function importEdge(data) {

    // Validating
    var _validateSerializedEd = (0, _serialization.validateSerializedEdge)(data);

    var valid = _validateSerializedEd.valid;
    var reason = _validateSerializedEd.reason;


    if (!valid) {
      if (reason === 'not-object') throw new _errors.InvalidArgumentsGraphError('Graph.importEdge: invalid serialized edge. A serialized edge should be a plain object with at least a "source" & "target" property.');
      if (reason === 'no-source') throw new _errors.InvalidArgumentsGraphError('Graph.importEdge: missing souce.');
      if (reason === 'no-target') throw new _errors.InvalidArgumentsGraphError('Graph.importEdge: missing target');
      if (reason === 'invalid-attributes') throw new _errors.InvalidArgumentsGraphError('Graph.importEdge: invalid attributes. Attributes should be a plain object, null or omitted.');
      if (reason === 'invalid-undirected') throw new _errors.InvalidArgumentsGraphError('Graph.importEdge: invalid undirected. Undirected should be boolean or omitted.');
    }

    // Adding the edge
    var source = data.source;
    var target = data.target;
    var _data$attributes2 = data.attributes;
    var attributes = _data$attributes2 === undefined ? {} : _data$attributes2;
    var _data$undirected2 = data.undirected;
    var undirected = _data$undirected2 === undefined ? false : _data$undirected2;


    var method = void 0;

    if ('key' in data) {
      method = undirected ? this.addUndirectedEdgeWithKey : this.addEdgeWithKey;

      method.call(this, data.key, source, target, attributes);
    } else {
      method = undirected ? this.addUndirectedEdge : this.addDirectedEdge;

      method.call(this, source, target, attributes);
    }

    return this;
  };

  /**
   * Method used to import serialized nodes.
   *
   * @param  {array} nodes - The serialized nodes.
   * @return {Graph}       - Returns itself for chaining.
   */


  Graph.prototype.importNodes = function importNodes(nodes) {
    if (!Array.isArray(nodes)) throw new _errors.InvalidArgumentsGraphError('Graph.importNodes: invalid argument. Expecting an array.');

    for (var i = 0, l = nodes.length; i < l; i++) {
      this.importNode(nodes[i]);
    }return this;
  };

  /**
   * Method used to import serialized edges.
   *
   * @param  {array} edges - The serialized edges.
   * @return {Graph}       - Returns itself for chaining.
   */


  Graph.prototype.importEdges = function importEdges(edges) {
    if (!Array.isArray(edges)) throw new _errors.InvalidArgumentsGraphError('Graph.importEdges: invalid argument. Expecting an array.');

    for (var i = 0, l = edges.length; i < l; i++) {
      this.importEdge(edges[i]);
    }return this;
  };

  /**
   * Method used to import a serialized graph.
   *
   * @param  {object|Graph} data - The serialized graph.
   * @return {Graph}             - Returns itself for chaining.
   */


  Graph.prototype.import = function _import(data) {

    // Importing a Graph instance
    if ((0, _utils.isGraph)(data)) {

      this.import(data.export());
      return this;
    }

    // Importing a serialized graph
    if (!(0, _utils.isPlainObject)(data) || !data.nodes) throw new _errors.InvalidArgumentsGraphError('Graph.import: invalid argument. Expecting an object with at least a "nodes" property or, alternatively, a Graph instance.');

    this.importNodes(data.nodes);

    if (data.edges) this.importEdges(data.edges);

    return this;
  };

  /**
   * Method returning an empty copy of the graph, i.e. a graph without nodes
   * & edges but with the exact same options.
   *
   * @return {Graph} - The empty copy.
   */


  Graph.prototype.emptyCopy = function emptyCopy() {
    return new Graph(null, this._options);
  };

  /**
   * Method returning an exact copy of the graph.
   *
   * @return {Graph} - The copy.
   */


  Graph.prototype.copy = function copy() {
    return new Graph(this, this._options);
  };

  /**---------------------------------------------------------------------------
   * Indexes-related methods
   **---------------------------------------------------------------------------
   */

  /**
   * Method computing the desired index.
   *
   * @param  {string} name - Name of the index to compute.
   * @return {Graph}       - Returns itself for chaining.
   *
   * @throw  {Error} - Will throw if the index doesn't exist.
   */


  Graph.prototype.computeIndex = function computeIndex(name) {
    var _this9 = this;

    if (!_indices.INDICES.has(name)) throw new _errors.InvalidArgumentsGraphError('Graph.computeIndex: unknown "' + name + '" index.');

    if (name === 'structure') {
      var index = this._indices.structure;

      if (index.computed) return this;

      index.computed = true;

      if (this.map) {
        this._edges.forEach(function (data, edge) {
          return _this9._updateIndex(name, edge, data);
        });
      } else {
        for (var edge in this._edges) {
          this._updateIndex(name, edge, this._edges[edge]);
        }
      }
    }

    return this;
  };

  /**
   * Method updating the desired index.
   *
   * @param  {string} name      - Name of the index to compute.
   * @param  {mixed}  [...args] - Additional arguments.
   * @return {Graph}            - Returns itself for chaining.
   *
   * @throw  {Error} - Will throw if the index doesn't exist.
   */


  Graph.prototype._updateIndex = function _updateIndex(name) {
    if (!_indices.INDICES.has(name)) throw new _errors.InvalidArgumentsGraphError('Graph._updateIndex: unknown "' + name + '" index.');

    if (name === 'structure') {
      var index = this._indices.structure;

      if (!index.computed) return this;

      for (var _len = arguments.length, args = Array(_len > 1 ? _len - 1 : 0), _key = 1; _key < _len; _key++) {
        args[_key - 1] = arguments[_key];
      }

      var edge = args[0];
      var data = args[1];


      (0, _indices.updateStructureIndex)(this, edge, data);
    }

    return this;
  };

  /**
   * Method used to clear an edge from the desired index to clear memory.
   *
   * @param  {string} name - Name of the index to update.
   * @param  {any}    edge - Target edge.
   * @param  {object} data - Former attached data.
   * @return {Graph}       - Returns itself for chaining.
   *
   * @throw  {Error} - Will throw if the index doesn't exist.
   */


  Graph.prototype._clearEdgeFromIndex = function _clearEdgeFromIndex(name, edge, data) {
    if (!_indices.INDICES.has(name)) throw new _errors.InvalidArgumentsGraphError('Graph._clearEdgeFromIndex: unknown "' + name + '" index.');

    if (name === 'structure') {
      var index = this._indices.structure;

      if (!index.computed) return this;

      (0, _indices.clearEdgeFromStructureIndex)(this, edge, data);
    }

    return this;
  };

  /**
   * Method used to clear the desired index to clear memory.
   *
   * @param  {string} name - Name of the index to compute.
   * @return {Graph}       - Returns itself for chaining.
   *
   * @throw  {Error} - Will throw if the index doesn't exist.
   */


  Graph.prototype.clearIndex = function clearIndex(name) {
    if (!_indices.INDICES.has(name)) throw new _errors.InvalidArgumentsGraphError('Graph.clearIndex: unknown "' + name + '" index.');

    if (name === 'structure') {
      var index = this._indices.structure;

      if (!index.computed) return this;

      (0, _indices.clearStructureIndex)(this);
      index.computed = false;
    }

    return this;
  };

  /**---------------------------------------------------------------------------
   * Known methods
   **---------------------------------------------------------------------------
   */

  /**
   * Method used by JavaScript to perform JSON serialization.
   *
   * @return {object} - The serialized graph.
   */


  Graph.prototype.toJSON = function toJSON() {
    return this.export();
  };

  /**
   * Method used to perform string coercion and returning useful information
   * about the Graph instance.
   *
   * @return {string} - String representation of the graph.
   */


  Graph.prototype.toString = function toString() {
    var pluralOrder = this.order > 1 || this.order === 0,
        pluralSize = this.size > 1 || this.size === 0;

    return 'Graph<' + (0, _utils.prettyPrint)(this.order) + ' node' + (pluralOrder ? 's' : '') + ', ' + (0, _utils.prettyPrint)(this.size) + ' edge' + (pluralSize ? 's' : '') + '>';
  };

  /**
   * Method used internally by node's console to display a custom object.
   *
   * @return {object} - Formatted object representation of the graph.
   */


  Graph.prototype.inspect = function inspect() {
    var nodes = void 0,
        edges = void 0;

    if (this.map) {
      nodes = new Map();
      this._nodes.forEach(function (value, key) {
        var attributes = value.attributes;

        nodes.set(key, Object.keys(attributes).length ? attributes : '<empty>');
      });

      edges = [];
      this._edges.forEach(function (value, key) {

        var formatted = [key, value.source, value.undirected ? '<->' : '->', value.target];

        if (Object.keys(value.attributes).length) formatted.push(value.attributes);

        edges.push(formatted);
      });
    } else {
      nodes = {};
      edges = [];

      for (var k in this._nodes) {
        var attributes = this._nodes[k].attributes;
        nodes[k] = Object.keys(attributes).length ? attributes : '<empty>';
      }

      for (var _k in this._edges) {
        var value = this._edges[_k];

        var formatted = [_k, value.source, value.undirected ? '<->' : '->', value.target];

        if (Object.keys(value.attributes).length) formatted.push(value.attributes);

        edges.push(formatted);
      }
    }

    var dummy = {};

    for (var _k2 in this) {
      if (this.hasOwnProperty(_k2) && !EMITTER_PROPS.has(_k2)) dummy[_k2] = this[_k2];
    }

    dummy.nodes = nodes;
    dummy.edges = edges;

    (0, _utils.privateProperty)(dummy, 'constructor', this.constructor);

    return dummy;
  };

  /**
   * Method used to inspect the internal data structure holding nodes & edges.
   *
   * @return {object} - The internals to show.
   */


  Graph.prototype.internals = function internals() {
    return {
      nodes: this._nodes,
      edges: this._edges
    };
  };

  return Graph;
}(_events.EventEmitter);

/**
 * Attaching iteration methods to the prototype.
 *
 * Here, we create iteration methods for every kind of iteration by attaching
 * them to the Graph class prototype rather than writing a lot of custom methods
 * one by one.
 */

/**
 * Attributes-related.
 */


exports.default = Graph;
(0, _attributes.attachAttributesMethods)(Graph);

/**
 * Edge iteration-related.
 */
(0, _edges.attachEdgeIterationMethods)(Graph);

/**
 * Neighbor iteration-related.
 */
(0, _neighbors.attachNeighborIterationMethods)(Graph);